import {readFile, writeFile} from 'fs/promises';
import {join} from 'path';
import {DataSource as DataSourceInterface, ResourceIdConfig} from '../data-source';
import {Resource} from '../..';
import * as v from 'valibot';

declare module '../..' {


	interface BaseResourceState {
		idAttr: string;
		idConfig: ResourceIdConfig<v.BaseSchema>
	}

	interface Resource<
		Schema extends v.BaseSchema = v.BaseSchema,
		CurrentName extends string = string,
		CurrentRouteName extends string = string,
		CurrentIdAttr extends string = string,
		IdSchema extends v.BaseSchema = v.BaseSchema
	> {
		dataSource?: DataSourceInterface;
		id<NewIdAttr extends CurrentIdAttr, TheIdSchema extends IdSchema>(
			newIdAttr: NewIdAttr,
			params: ResourceIdConfig<TheIdSchema>
		): Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, TheIdSchema>;
		fullText(fullTextAttr: string): this;
	}
}

export class DataSource<T extends Record<string, string>> implements DataSourceInterface<T> {
	private path?: string;

	private resource?: Resource;

	data: T[] = [];

	constructor(private readonly baseDir = '') {
		// noop
	}

	prepareResource<
		Schema extends v.BaseSchema = v.BaseSchema,
		CurrentName extends string = string,
		CurrentRouteName extends string = string
	>(resource: Resource<Schema, CurrentName, CurrentRouteName>) {
		this.path = join(this.baseDir, `${resource.state.routeName}.jsonl`);
		resource.dataSource = resource.dataSource ?? this;
		const originalResourceId = resource.id;
		resource.id = <NewIdAttr extends string, NewIdSchema extends v.BaseSchema>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) => {
			originalResourceId(newIdAttr, params);
			return resource as Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, NewIdSchema>;
		};
		this.resource = resource;
	}

	async initialize() {
		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		try {
			const fileContents = await readFile(this.path, 'utf-8');
			const lines = fileContents.split('\n');
			this.data = lines.filter((l) => l.trim().length > 0).map((l) => JSON.parse(l));
		} catch (err) {
			await writeFile(this.path, '');
		}
	}

	async getTotalCount() {
		return this.data.length;
	}

	async getMultiple() {
		return [...this.data];
	}

	async newId() {
		const idConfig = this.resource?.state.shared.get('idConfig') as ResourceIdConfig<any>;
		if (typeof idConfig === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		const theNewId = await idConfig.generationStrategy(this);
		return theNewId as string;
	}

	async getById(idSerialized: string) {
		if (typeof this.resource === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		const theIdAttr = this.resource.state.shared.get('idAttr');
		if (typeof theIdAttr === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const idAttr = theIdAttr as string;
		const theIdConfigRaw = this.resource.state.shared.get('idConfig');
		if (typeof theIdConfigRaw === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const theIdConfig = theIdConfigRaw as ResourceIdConfig<any>;

		const id = theIdConfig.deserialize(idSerialized);
		const foundData = this.data.find((s) => s[idAttr] === id);

		if (foundData) {
			return {
				...foundData
			};
		}

		return null;
	}

	async create(data: T) {
		if (typeof this.resource === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		const theIdAttr = this.resource.state.shared.get('idAttr');
		if (typeof theIdAttr === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const idAttr = theIdAttr as string;
		const theIdConfigRaw = this.resource.state.shared.get('idConfig');
		if (typeof theIdConfigRaw === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const theIdConfig = theIdConfigRaw as ResourceIdConfig<any>;

		const newData = {
			...data
		} as Record<string, unknown>;

		if (idAttr in newData) {
			newData[idAttr] = theIdConfig.deserialize(newData[idAttr] as string);
		}

		const newCollection = [
			...this.data,
			newData
		];

		await writeFile(this.path, newCollection.map((d) => JSON.stringify(d)).join('\n'));

		return data as T;
	}

	async delete(idSerialized: string) {
		if (typeof this.resource === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		const theIdAttr = this.resource.state.shared.get('idAttr');
		if (typeof theIdAttr === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const idAttr = theIdAttr as string;
		const theIdConfigRaw = this.resource.state.shared.get('idConfig');
		if (typeof theIdConfigRaw === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const theIdConfig = theIdConfigRaw as ResourceIdConfig<any>;

		const oldDataLength = this.data.length;

		const id = theIdConfig.deserialize(idSerialized);
		const newData = this.data.filter((s) => !(s[idAttr] === id));

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

		return oldDataLength !== newData.length;
	}

	async emplace(idSerialized: string, dataWithId: T) {
		if (typeof this.resource === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		const theIdAttr = this.resource.state.shared.get('idAttr');
		if (typeof theIdAttr === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const idAttr = theIdAttr as string;
		const theIdConfigRaw = this.resource.state.shared.get('idConfig');
		if (typeof theIdConfigRaw === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const theIdConfig = theIdConfigRaw as ResourceIdConfig<any>;

		const existing = await this.getById(idSerialized);
		const id = theIdConfig.deserialize(idSerialized);
		const { [idAttr]: idFromResource, ...data } = dataWithId;
		const dataToEmplace = {
			...data,
			[idAttr]: id,
		} as T;

		if (existing) {
			const newData = this.data.map((d) => {
				if (d[idAttr] === id) {
					return dataToEmplace;
				}

				return d;
			});

			await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

			return [dataToEmplace, false] as [T, boolean];
		}

		const newData = await this.create(dataToEmplace);
		return [newData, true] as [T, boolean];
	}

	async patch(idSerialized: string, data: Partial<T>) {
		if (typeof this.resource === 'undefined') {
			throw new Error('Resource not prepared.');
		}

		if (typeof this.path !== 'string') {
			throw new Error('Resource not prepared.');
		}

		const theIdAttr = this.resource.state.shared.get('idAttr');
		if (typeof theIdAttr === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const idAttr = theIdAttr as string;
		const theIdConfigRaw = this.resource.state.shared.get('idConfig');
		if (typeof theIdConfigRaw === 'undefined') {
			throw new Error('Resource not prepared.');
		}
		const theIdConfig = theIdConfigRaw as ResourceIdConfig<any>;

		const existing = await this.getById(idSerialized);
		if (!existing) {
			return null;
		}

		const newItem = {
			...existing,
			...data,
		}

		const id = theIdConfig.deserialize(idSerialized);
		const newData = this.data.map((d) => {
			if (d[idAttr] === id) {
				return newItem;
			}

			return d;
		});

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));
		return newItem as T;
	}
}
