import http from 'http';
import {BackendState, RequestContext} from '../common';
import {Language, Resource, LanguageStatusMessageMap} from '../../common';
import https from 'https';
import {constants} from 'http2';
import * as v from 'valibot';
import {
	handleCreateItem,
	handleDeleteItem,
	handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handlePatchItem,
} from './handlers';
import {getBody} from './utils';
import {decorateRequestWithBackend} from './decorators/backend';
import {decorateRequestWithMethod} from './decorators/method';
import {decorateRequestWithUrl} from './decorators/url';

declare module '../common' {
	interface RequestContext extends http.IncomingMessage {
		body?: unknown;
	}
}

export interface Response {
	statusCode: number;

	statusMessage?: keyof LanguageStatusMessageMap;

	headers?: Record<string, string>;
}

interface ResponseParams {
	statusCode: Response['statusCode'];
	statusMessage?: Response['statusMessage'];
	headers?: Response['headers'];
}

export class MiddlewareError extends Error {}

interface PlainResponseParams<T = unknown> extends ResponseParams {
	body?: T;
}

interface HttpMiddlewareErrorParams<T = unknown> extends Omit<PlainResponseParams<T>, 'statusMessage'> {
	cause?: unknown
}

export class PlainResponse<T = unknown> implements Response {
	readonly statusCode: Response['statusCode'];

	readonly statusMessage?: keyof LanguageStatusMessageMap;

	readonly headers: Response['headers'];

	readonly body?: T;

	constructor(args: PlainResponseParams<T>) {
		this.statusCode = args.statusCode;
		this.statusMessage = args.statusMessage;
		this.headers = args.headers;
		this.body = args.body;
	}
}

export class HttpMiddlewareError extends MiddlewareError {
	readonly response: PlainResponse;

	constructor(statusMessage: keyof Language['statusMessages'], params: HttpMiddlewareErrorParams) {
		super(statusMessage, { cause: params.cause });
		this.response = new PlainResponse({
			...params,
			statusMessage,
		});
	}
}

export interface CreateServerParams {
	basePath?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
	// CQRS
	streamResponses?: boolean;
}

type ResourceRequestContext = Omit<RequestContext, 'resource'> & Required<Pick<RequestContext, 'resource'>>;

export interface Middleware<Req extends ResourceRequestContext = ResourceRequestContext> {
	(req: Req): undefined | Response | Promise<undefined | Response>;
}

type Method = 'GET' | 'POST' | 'PUT' | 'PATCH' | 'DELETE';

interface AllowedMiddlewareSpecification<Schema extends v.BaseSchema = v.BaseSchema> {
	method: Method;
	middleware: Middleware;
	constructBodySchema?: (resource: Resource<Schema>, resourceId?: string) => v.BaseSchema;
	allowed: (resource: Resource<Schema>) => boolean;
}

const constructPostSchema = <T extends v.BaseSchema>(resource: Resource<T>) => {
	return resource.schema;
};

const constructPutSchema = <T extends v.BaseSchema>(resource: Resource<T>, mainResourceId?: string) => {
	if (typeof mainResourceId === 'undefined') {
		return resource.schema;
	}

	const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;
	const idAttr = resource.state.shared.get('idAttr') as string;
	const idConfig = resource.state.shared.get('idConfig') as any;
	return (
		schema.type === 'object'
			? v.merge([
				schema as v.ObjectSchema<any>,
				v.object({
					[idAttr]: v.transform(
						v.any(),
						input => idConfig!.serialize(input),
						v.literal(mainResourceId)
					)
				})
			])
			: schema
	);
};

const constructPatchSchema = <T extends v.BaseSchema>(resource: Resource<T>) => {
	const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema;
	return (
		schema.type === 'object'
			? v.partial(
				schema as v.ObjectSchema<any>,
				(schema as v.ObjectSchema<any>).rest,
				(schema as v.ObjectSchema<any>).pipe
			)
			: schema
	);
};

// TODO add a way to define custom middlewares
const defaultCollectionMiddlewares: AllowedMiddlewareSpecification[] = [
	{
		method: 'GET',
		middleware: handleGetCollection,
		allowed: (resource) => resource.state.canFetchCollection,
	},
	{
		method: 'POST',
		middleware: handleCreateItem,
		allowed: (resource) => resource.state.canCreate,
		constructBodySchema: constructPostSchema,
	},
];

const defaultItemMiddlewares: AllowedMiddlewareSpecification[] = [
	{
		method: 'GET',
		middleware: handleGetItem,
		allowed: (resource) => resource.state.canFetchItem,
	},
	{
		method: 'PUT',
		middleware: handleEmplaceItem,
		constructBodySchema: constructPutSchema,
		allowed: (resource) => resource.state.canEmplace,
	},
	{
		method: 'PATCH',
		middleware: handlePatchItem,
		constructBodySchema: constructPatchSchema,
		allowed: (resource) => resource.state.canPatch,
	},
	{
		method: 'DELETE',
		middleware: handleDeleteItem,
		allowed: (resource) => resource.state.canDelete,
	},
];

export const createServer = (backendState: BackendState, serverParams = {} as CreateServerParams) => {
	const isHttps = 'key' in serverParams && 'cert' in serverParams;

	const server = isHttps
		? https.createServer({
			key: serverParams.key,
			cert: serverParams.cert,
			requestTimeout: serverParams.requestTimeout,
		})
		: http.createServer({
			requestTimeout: serverParams.requestTimeout,
		});

	const defaultRequestDecorators = [
		decorateRequestWithMethod,
		decorateRequestWithUrl(serverParams),
		decorateRequestWithBackend(backendState),
	];

	const isTextMediaType = (mediaType: string) => (
		mediaType.startsWith('text/')
		|| [
			'application/json',
			'application/xml'
		].includes(mediaType)
	);

	const handleMiddlewares = async (currentHandlerState: Awaited<ReturnType<Middleware>>, currentMiddleware: AllowedMiddlewareSpecification, req: ResourceRequestContext) => {
		const { method: middlewareMethod, middleware, constructBodySchema} = currentMiddleware;
		const effectiveMethod = req.method === 'HEAD' ? 'GET' : req.method;

		if (effectiveMethod !== middlewareMethod) {
			return currentHandlerState;
		}

		if (typeof currentHandlerState !== 'undefined') {
			return currentHandlerState;
		}

		if (typeof constructBodySchema === 'function') {
			const bodySchema = constructBodySchema(req.resource, req.resourceId);
			const contentTypeHeader = req.headers['content-type'] ?? 'application/octet-stream';
			const fragments = contentTypeHeader.replace(/\s+/g, ' ').split(';');
			const mediaType = fragments[0];
			const charsetParam = (
				fragments
					.map((s) => s.trim())
					.find((f) => f.startsWith('charset='))

				?? (
					isTextMediaType(mediaType)
						? 'charset=utf-8'
						: 'charset=binary'
				)
			);
			const [_charsetKey, charsetRaw] = charsetParam.split('=').map((s) => s.trim());
			const charset = (
				(
					(charsetRaw.startsWith('"') && charsetRaw.endsWith('"'))
					|| (charsetRaw.startsWith("'") && charsetRaw.endsWith("'"))
				)
					? charsetRaw.slice(1, -1).trim()
					: charsetRaw.trim()
			)
			const theBodyBuffer = await getBody(req);
			const encodingPair = req.backend.app.charsets.get(charset);
			if (typeof encodingPair === 'undefined') {
				throw new HttpMiddlewareError('unableToDecodeResource', {
					statusCode: constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE,
				});
			}
			const deserializerPair = req.backend.app.mediaTypes.get(mediaType);
			if (typeof deserializerPair === 'undefined') {
				throw new HttpMiddlewareError('unableToDeserializeResource', {
					statusCode: constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE,
				});
			}
			const theBodyStr = encodingPair.decode(theBodyBuffer);
			const theBody = deserializerPair.deserialize(theBodyStr);
			try {
				req.body = await v.parseAsync(bodySchema, theBody, {abortEarly: false, abortPipeEarly: false});
			} catch (errRaw) {
				const err = errRaw as v.ValiError;
				// todo use error message key for each method
				// TODO better error reporting, localizable messages
				// TODO handle error handlers' errors
				if (Array.isArray(err.issues)) {
					throw new HttpMiddlewareError('invalidResource', {
						statusCode: constants.HTTP_STATUS_BAD_REQUEST,
						body: err.issues.map((i) => (
							`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
						)),
					});
				}
			}
		}

		const result = await middleware(req);

		// HEAD is just GET without the response body
		if (req.method === 'HEAD' && result instanceof PlainResponse) {
			const { body: _, ...etcResult } = result;

			return new PlainResponse(etcResult);
		}

		return result;
	};

	const processRequest = (middlewares: AllowedMiddlewareSpecification[]) => async (req: ResourceRequestContext) => {
		if (req.url === '/' || req.url === '') {
			return handleGetRoot(req);
		}

		const { resource } = req;
		if (typeof resource === 'undefined') {
			throw new HttpMiddlewareError('resourceNotFound', {
				statusCode: constants.HTTP_STATUS_NOT_FOUND,
			});
		}

		if (typeof resource.dataSource === 'undefined') {
			throw new HttpMiddlewareError('unableToInitializeResourceDataSource', {
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			});
		}

		try {
			await resource.dataSource.initialize();
		} catch (cause) {
			throw new HttpMiddlewareError(
				'unableToInitializeResourceDataSource',
				{
					cause,
					statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
				}
			);
		}

		const middlewareResponse = await middlewares.reduce<ReturnType<Middleware>>(
			async (currentHandlerStatePromise, currentMiddleware) => {
				const currentHandlerState = await currentHandlerStatePromise;
				return await handleMiddlewares(currentHandlerState, currentMiddleware, req);
			},
			Promise.resolve<ReturnType<Middleware>>(undefined)
		) as Awaited<ReturnType<Middleware>>;

		if (typeof middlewareResponse === 'undefined') {
			throw new HttpMiddlewareError('resourceNotFound', {
				statusCode: constants.HTTP_STATUS_NOT_FOUND
			});
		}

		return middlewareResponse as Awaited<ReturnType<Middleware>>
	};

	const decorateRequest = async (reqRaw: http.IncomingMessage) => {
		// TODO custom decorators
		const effectiveRequestDecorators = defaultRequestDecorators;
		return await effectiveRequestDecorators.reduce(
			async (resultRequestPromise, decorator) => {
				const resultRequest = await resultRequestPromise;

				return await decorator(resultRequest);
			},
			Promise.resolve(reqRaw as RequestContext)
		);
	};

	const handleRequest = async (reqRaw: RequestContext, res: http.ServerResponse<RequestContext>) => {
		const plainReq = await decorateRequest(reqRaw); // TODO add type safety here, put handleGetRoot as its own middleware as it does not concern over any resource
		if (typeof plainReq.resource !== 'undefined') {
			const resourceReq = plainReq as ResourceRequestContext;
			const effectiveMiddlewares = (
				typeof resourceReq.resourceId === 'string'
					? defaultItemMiddlewares
					: defaultCollectionMiddlewares
			);
			const middlewares = effectiveMiddlewares.filter((m) => m.allowed(resourceReq.resource));
			const processRequestFn = processRequest(middlewares);
			let middlewareState: Response;
			try {
				middlewareState = await processRequestFn(resourceReq) as any; // TODO fix this
			} catch (processRequestErrRaw) {
				const finalErr = processRequestErrRaw as HttpMiddlewareError;
				const headers = finalErr.response.headers ?? {};
				let encoded: Buffer | undefined;
				let serialized;
				try {
					serialized = typeof finalErr.response.body !== 'undefined' ? resourceReq.backend.cn.mediaType.serialize(finalErr.response.body) : undefined;
				} catch (cause) {
					res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
					res.end();
					return;
				}

				try {
					encoded = typeof serialized !== 'undefined' ? resourceReq.backend.cn.charset.encode(serialized) : undefined;
				} catch (cause) {
					res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR);
					res.end();
				}

				headers['Content-Type'] = [
					resourceReq.backend.cn.mediaType.name,
					`charset=${resourceReq.backend.cn.charset.name}`,
				].join('; ');

				const statusMessageKey = finalErr.response.statusMessage ? resourceReq.backend.cn.language.statusMessages[finalErr.response.statusMessage] : undefined;
				res.statusMessage = statusMessageKey?.replace(/\$RESOURCE/g, resourceReq.resource!.state.itemName) ?? '';
				res.writeHead(finalErr.response.statusCode, headers);
				if (typeof encoded !== 'undefined') {
					res.end(encoded);
					return;
				}
				res.end();
				return;
			}

			//  TODO extract to separate function
			const headers: Record<string, string> = {
				...(
					middlewareState.headers ?? {}
				),
				'Content-Language': resourceReq.cn.language.name,
			};
			if (middlewareState instanceof http.ServerResponse) {
				// TODO streaming responses
				middlewareState.writeHead(constants.HTTP_STATUS_ACCEPTED, headers);
				return;
			}
			if (middlewareState instanceof PlainResponse) {
				let encoded: Buffer | undefined;
				if (typeof middlewareState.body !== 'undefined') {
					let serialized;
					try {
						serialized = resourceReq.cn.mediaType.serialize(middlewareState.body);
					} catch (cause) {
						const headers: Record<string, string> = {
							'Content-Language': resourceReq.backend.cn.language.name,
						};
						if (resourceReq.method === 'POST') {
							headers['Accept-Post'] = Array.from(resourceReq.backend.app.mediaTypes.keys()).join(',');
						} else if (resourceReq.method === 'PATCH') {
							headers['Accept-Patch'] = Array.from(resourceReq.backend.app.mediaTypes.keys()).map((m) => {
								const [mimeType, mimeSubtype] = m.split('/');

								// TODO accept only patch document type from request
								// TODO implement Vary header (which headers influenced the request)
								// TODO implement OPTIONS method for determining the accepted media types and languages
								// TODO configure strict and lax accept behavior for content negotiation
								return `${mimeType}/merge-patch+${mimeSubtype}`;
							}).join(',');
						}
						res.statusMessage = resourceReq.backend.cn.language.statusMessages['unableToSerializeResponse']?.replace(
							/\$RESOURCE/g,
							resourceReq.resource!.state.itemName) ?? '';
						res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, headers);
						res.end();
						return;
					}

					try {
						encoded = resourceReq.cn.charset.encode(serialized);
					} catch (cause) {
						res.statusMessage = resourceReq.backend.cn.language.statusMessages['unableToEncodeResponse']?.replace(/\$RESOURCE/g,
							resourceReq.resource!.state.itemName) ?? '';
						res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
							'Content-Language': resourceReq.backend.cn.language.name,
						});
						res.end();
						return;
					}

					headers['Content-Type'] = [
						resourceReq.cn.mediaType.name,
						`charset=${resourceReq.cn.charset.name}`,
					].join('; ');
				}

				const statusMessageKey = middlewareState.statusMessage ? resourceReq.cn.language.statusMessages[middlewareState.statusMessage] : undefined;
				res.statusMessage = statusMessageKey?.replace(/\$RESOURCE/g, resourceReq.resource!.state.itemName) ?? '';
				res.writeHead(middlewareState.statusCode, headers);
				if (typeof encoded !== 'undefined') {
					res.end(encoded);
					return;
				}
				res.end();
				return;
			}

			if (middlewares.length > 0) {
				res.statusMessage = resourceReq.backend.cn.language.statusMessages.methodNotAllowed.replace(/\$RESOURCE/g,
					resourceReq.resource!.state.itemName) ?? '';
				res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
					Allow: middlewares.map((m) => m.method).join(', '),
					'Content-Language': resourceReq.backend.cn.language.name,
				});
				res.end();
				return;
			}
			res.statusMessage = resourceReq.backend.cn.language.statusMessages.urlNotFound.replace(/\$RESOURCE/g,
				resourceReq.resource!.state.itemName) ?? '';
			res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
				'Content-Language': resourceReq.backend.cn.language.name,
			});
			res.end();
			return;
		}

		throw new Error('Not implemented');
	};

	server.on('request', handleRequest);

	return server;
}
