import {Resource} from './resource';
import {FALLBACK_LANGUAGE, Language} from './language';
import {FALLBACK_MEDIA_TYPE, MediaType} from './media-type';
import {Charset, FALLBACK_CHARSET} from './charset';
import * as v from 'valibot';
import {BackendBuilder, createBackend, CreateBackendParams} from '../backend';
import {ClientBuilder, createClient, CreateClientParams} from '../client';

type ApplicationMap<T extends { name: string }> = Map<T['name'], T>;

export interface ApplicationState {
	name: string;
	resources: Set<Resource<any>>;
	languages: ApplicationMap<Language>;
	mediaTypes: ApplicationMap<MediaType>;
	charsets: ApplicationMap<Charset>;
}

export interface ApplicationParams {
	name: string;
}

export interface ApplicationBuilder {
	mediaType(mediaType: MediaType): this;
	language(language: Language): this;
	charset(charset: Charset): this;
	resource<
		Schema extends v.BaseSchema,
		CurrentItemName extends string = string,
		CurrentRouteName extends string = string
	>(resRaw: Resource<Schema, CurrentItemName, CurrentRouteName>): this;
	createBackend(params: Omit<CreateBackendParams, 'app'>): BackendBuilder;
	createClient(params: Omit<CreateClientParams, 'app'>): ClientBuilder;
}

export const application = (appParams: ApplicationParams): ApplicationBuilder => {
	const appState: ApplicationState = {
		name: appParams.name,
		resources: new Set<Resource<any>>(),
		languages: new Map<Language['name'], Language>([
			[FALLBACK_LANGUAGE.name, FALLBACK_LANGUAGE],
		]),
		mediaTypes: new Map<MediaType['name'], MediaType>([
			[FALLBACK_MEDIA_TYPE.name, FALLBACK_MEDIA_TYPE],
		]),
		charsets: new Map<Charset['name'], Charset>([
			[FALLBACK_CHARSET.name, FALLBACK_CHARSET],
		]),
	};

	return {
		mediaType(mediaType: MediaType) {
			appState.mediaTypes.set(mediaType.name, mediaType);
			return this;
		},
		charset(charset: Charset) {
			appState.charsets.set(charset.name, charset);
			return this;
		},
		language(language: Language) {
			appState.languages.set(language.name, language);
			return this;
		},
		resource<T extends v.BaseSchema>(resRaw: Resource<T>) {
			appState.resources.add(resRaw);
			return this;
		},
		createBackend(params: Omit<CreateBackendParams, 'app'>) {
			return createBackend({
				...params,
				app: appState
			});
		},
		createClient(params: Omit<CreateClientParams, 'app'>) {
			return createClient({
				...params,
				app: appState
			});
		},
	};
};
