import * as v from 'valibot';

export interface ResourceState<
	ItemName extends string = string,
	RouteName extends string = string
> {
	shared: Map<string, unknown>;
	itemName: ItemName;
	routeName: RouteName;
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: boolean;
	canEmplace: boolean;
	canDelete: boolean;
}

export interface Resource<
	Schema extends v.BaseSchema = v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string
> {
	schema: Schema;
	state: ResourceState<CurrentName, CurrentRouteName>;
	name<NewName extends CurrentName>(n: NewName): Resource<Schema, NewName, CurrentRouteName>;
	route<NewRouteName extends CurrentRouteName>(n: NewRouteName): Resource<Schema, CurrentName, NewRouteName>;
	canFetchCollection(b?: boolean): this;
	canFetchItem(b?: boolean): this;
	canCreate(b?: boolean): this;
	canPatch(b?: boolean): this;
	canEmplace(b?: boolean): this;
	canDelete(b?: boolean): this;
}

export const resource = <
	Schema extends v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string
>(schema: Schema): Resource<Schema, CurrentName, CurrentRouteName> => {
	const resourceState = {
		shared: new Map(),
		canCreate: false,
		canFetchCollection: false,
		canFetchItem: false,
		canPatch: false,
		canEmplace: false,
		canDelete: false,
	} as ResourceState<CurrentName, CurrentRouteName>;

	return {
		get state(): ResourceState<CurrentName, CurrentRouteName> {
			return Object.freeze({
				...resourceState
			}) as unknown as ResourceState<CurrentName, CurrentRouteName>;
		},
		canFetchCollection(b = true) {
			resourceState.canFetchCollection = b;
			return this;
		},
		canFetchItem(b = true) {
			resourceState.canFetchItem = b;
			return this;
		},
		canCreate(b = true) {
			resourceState.canCreate = b;
			return this;
		},
		canPatch(b = true) {
			resourceState.canPatch = b;
			return this;
		},
		canEmplace(b = true) {
			resourceState.canEmplace = b;
			return this;
		},
		canDelete(b = true) {
			resourceState.canDelete = b;
			return this;
		},
		id(idName, config) {
			resourceState.shared.set('idAttr', idName);
			resourceState.shared.set('idConfig', config);
			return this;
		},
		fullText(attrName) {
			const fullTextAttrs = (resourceState.shared.get('fullText') ?? new Set()) as Set<string>;
			fullTextAttrs.add(attrName);
			resourceState.shared.set('fullText', fullTextAttrs);
			return this;
		},
		name<NewName extends CurrentName>(n: NewName) {
			resourceState.itemName = n;
			return this as Resource<Schema, NewName, CurrentRouteName>;
		},
		route<NewRouteName extends CurrentRouteName>(n: NewRouteName) {
			resourceState.routeName = n;
			return this as Resource<Schema, CurrentName, NewRouteName>;
		},
		get itemName() {
			return resourceState.itemName;
		},
		get routeName() {
			return resourceState.routeName;
		},
		get schema() {
			return schema;
		},
	} as Resource<Schema, CurrentName, CurrentRouteName>;
};

export type ResourceType<R extends Resource> = v.Output<R['schema']>;
